'-------------------------------------------------
'		save and load png imagefile functions
'		Copyright (c) 2016 Masamichi Kirikae
'-------------------------------------------------

#include once "pnglib.bi"

namespace libpng

' invert_sign(a) a = -a
#define invert_sign(a) a = (not (a)) + 1

'--------------------------------------------------------------------
'	reverse uinteger32 endian
'--------------------------------------------------------------------
function bswap(byval n as uint32) as uint32

	dim p as ubyte ptr = cptr(ubyte ptr, @n)
	return(cast(uint32, p[0]) shl 24 or cast(uint32, p[1]) shl 16 or cast(uint32, p[2]) shl 8 or cast(uint32, p[3]))

end function

'--------------------------------------------------------------------
'	save png
'	only support 32bit depth (A8R8G8B8)
'--------------------------------------------------------------------
function save_png(byref filename as string, img_rgba() as ubyte, w as uint32, h as uint32, depth as uint32) as sint32

	dim f as sint32, ret as sint32
	dim ihdr as struct_ihdr = (bswap(w), bswap(h), 8, 6, 0, 0, 0)
	dim as uint32 ihdr_crc32 = crc32(ihdrcrc, cptr(ubyte ptr, @ihdr), sizeof(ihdr))
	
	dim as uint32 l, imgsize, idatsize, idat_crc32
	dim as uint32 x, y, n, n2
	dim imgdata(any) as ubyte
	dim idat(any) as ubyte

	l = w * 4 + 1
	imgsize = l * h
	idatsize = compressBound(imgsize)

	select case depth
		case 32
			redim imgdata(imgsize - 1) as ubyte
			redim idat(idatsize - 1) as ubyte
			g_width = w
			n = 0
			n2 = 0
			for y = 0 to h - 1
				imgdata(n2) = 0	' none filter
				n2+=1
				for x = 0 to w - 1
					imgdata(n2) = img_rgba(n): n+=1: n2+=1
					imgdata(n2) = img_rgba(n): n+=1: n2+=1
					imgdata(n2) = img_rgba(n): n+=1: n2+=1
					imgdata(n2) = img_rgba(n): n+=1: n2+=1
				next
			next

			if (compress2(@idat(0), @idatsize, @imgdata(0), imgsize, 9)) then
				return(PNG_ERROR_COMPRESS)
			endif
			idat_crc32 = crc32(idatcrc, @idat(0), idatsize)

			f = freefile
			open filename for output as #f
			if (Err > 0) then
				return(PNG_ERROR_OPEN) 
			endif

			put #f, , pngheader

			put #f, , bswap(13)
			put #f, , "IHDR"
			put #f, , ihdr
			put #f, , bswap(ihdr_crc32)

			put #f, , bswap(idatsize)
			put #f, , "IDAT"
			put #f, , idat(0), idatsize
			put #f, , bswap(idat_crc32)

			put #f, , bswap(0)
			put #f, , "IEND"
			put #f, , bswap(iendcrc)

			close #f
						
		case else
			return(PNG_ERROR_FORMAT)
	end select
	return(PNG_NO_ERROR)
end function

'--------------------------------------------------------------------
'	load png
'	only support 24/32bit depth (R8G8B8/A8R8G8B8)
'--------------------------------------------------------------------
function load_png(byref filename as string, image_rgba() as ubyte, byref imgw as sint32, byref imgh as sint32, byref bytepp as sint32) as sint32

	dim ihdr as struct_ihdr
	dim header as string * 8
	dim chunkname as string * 4, chunklength as sint32, chunkcrc as uint32
	dim dstfilename as string
	dim as sint32 f, size, rawlength, i, n, an
	dim dc as ubyte

	f = freefile
	open filename for binary access read as #f
	if (Err > 0) then return(0)

	get #f, 1, header
	if (header <> pngheader) then
		close #f
		return(0)
	endif

	get #f, , chunklength: chunklength = bswap(chunklength)
	get #f, , chunkname
	if (chunkname <> "IHDR") then
		close #f
		return(false)
	endif
	get #f, , ihdr
	imgw = bswap(ihdr.width)
	imgh = bswap(ihdr.height)
	if (imgw <= 0 or imgh <= 0) then
		close #f
		return(0)
	endif
	select case ihdr.colourtype
		case 0, 3:	bytepp = 1
		case 2: 		bytepp = 3
		case 4:		bytepp = 2
		case 6:		bytepp = 4
		case else
	end select

	if ((bytepp <> 3) and (bytepp <> 4)) then
		cprint "Error:"; bytepp; "bytes/pixel colour format is not supported."
		cprint "This version suports 24bit(R8G8B8) or 32bit(R8G8B8A8) colour format."
		close #f
		return(PNG_ERROR_FORMAT)
	endif
	
	size = imgh * (imgw * bytepp + 1) + 1
	if (size <= 1) then
		close #f
		return(PNG_ERROR_ANY)
	endif
	seek #f, seek(f) + 4
	
	' set shared variable
	g_width = imgw
	g_height = imgh
	g_bypp = bytepp	' actual bytes/pixel value
	g_bitdepth = 32	' 32 fixed (24bit colour conveted to 32bit colour automatically in this function)

	dim as sint32 actual_size, size32, x, y
	actual_size = imgw * imgh * 4
	size32 = imgh * (imgw * 4 + 1) + 1
	
	redim idat(size - 1) as ubyte
	redim buf(size - 1) as ubyte
	redim image(size32 - 1) as ubyte
	
	'dim imgp as ubyte ptr
	redim image_rgba(actual_size - 1) as ubyte
	
	rawlength = 0
	do
		get #f, ,chunklength:chunklength = bswap(chunklength)
		get #f, ,chunkname
		select case chunkname
			case "IDAT"
				get #f, ,idat(rawlength) ,chunklength
				get #f, ,chunkcrc
				rawlength += chunklength
			case "IEND"
				exit do
			case else
				for i = 1 to chunklength
					get #f, ,dc
				next
				get #f, ,chunkcrc
		end select
	loop
	close #f
		
	' convert RGB to RGBA. alpha value set to 0 (consider filtering).  
	if (bytepp = 3) then
		uncompress(@buf(0), @size, @idat(0), rawlength)
		n = 0: i = 0
		for y = 0 to imgh - 1
			image(n) = buf(i): n+=1: i+=1	' filter
			for x = 0 to imgw - 1
				image(n) = buf(i): n+=1: i+=1	' R
				image(n) = buf(i): n+=1: i+=1	' G
				image(n) = buf(i): n+=1: i+=1	' B
				image(n) = 0     : n+=1			' A
			next
		next
	else
		uncompress(@image(0), @size, @idat(0), rawlength)
	endif
		
	'	decode filter
	dim filter as sint32, c_rgba as rgba_data
	for y = 0 to imgh - 1
		n = y * ((imgw shl 2) + 1)
		filter = image(n): n+=1
		select case filter
			case 0	' none filter
				an = (y * imgw) shl 2	' means y * imgw * 4
				for x = 0 to imgw -1
					image_rgba(an) = image(n): an+=1: n+=1
					image_rgba(an) = image(n): an+=1: n+=1
					image_rgba(an) = image(n): an+=1: n+=1
					image_rgba(an) = image(n): an+=1: n+=1
					'pset_buffer(@image_rgba(0), x, y, c_rgba)
				next
			case 1	' sub filter
				decode_filter_sub(@image_rgba(0), @image(0), y)
			case 2	' up filter
				decode_filter_up(@image_rgba(0), @image(0), y)
			case 3	' average filter
				decode_filter_average(@image_rgba(0), @image(0), y)
			case 4	' paeth filter
				decode_filter_paeth(@image_rgba(0), @image(0), y)
			case else
				cprint "Unknown Filter:";filter
				return(PNG_ERROR_FORMAT)
		end select
	next
 
	' all pixel's alpha value must be 255 in RGB to RGBA conversion.
	if (bytepp = 3) then
		n = 0
		for y = 0 to imgh - 1
			for x = 0 to imgw - 1
				n+=3
				image_rgba(n) = &HFF
				n+=1
			next
		next
	endif
		
	return(PNG_NO_ERROR)
	
end function

sub decode_filter_sub(image_rgba as ubyte ptr, image as ubyte ptr, y as sint32)
	dim as sint32 x, n, an
	dim c as rgba_data

	n = y * (g_width * 4 + 1) + 1
	an = y * g_width * 4 
	for x = 0 to g_width - 1
		if (x > 0) then	
			' left pixel exist. reference is (decoded)left pixel.
			image_rgba[an] = cubyte(image_rgba[an-4] + cbyte(image[n])): an+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[an-4] + cbyte(image[n])): an+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[an-4] + cbyte(image[n])): an+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[an-4] + cbyte(image[n])): an+=1: n+=1
		else	
			' left pixel not exist. reference is original itself.
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
		end if
	next
end sub

sub decode_filter_up(image_rgba as ubyte ptr, image as ubyte ptr, y as sint32)
	dim as sint32 x, n, an, urn
	dim c as rgba_data
	
	an = y * g_width * 4
	if (y > 0) then	
		' above line exist. reference is above decoded pixel(d-pixel).
		n = y * (g_width * 4 + 1) + 1
		urn = (y - 1) * g_width * 4
		for x = 0 to g_width - 1
			image_rgba[an] = cubyte(image_rgba[urn] + cbyte(image[n])): an+=1: urn+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[urn] + cbyte(image[n])): an+=1: urn+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[urn] + cbyte(image[n])): an+=1: urn+=1: n+=1
			image_rgba[an] = cubyte(image_rgba[urn] + cbyte(image[n])): an+=1: urn+=1: n+=1
		next
	else	
		' above line not exist. reference is original itself.
		n = y * (g_width * 4 + 1) + 1
		for x = 0 to g_width - 1
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
			image_rgba[an] = image[n]: an+=1: n+=1
		next
	end if
end sub

sub decode_filter_average(image_rgba as ubyte ptr, image as ubyte ptr, y as sint32)
	dim as sint32 x, n, an, lrn, urn
	dim c as rgba_data

	an = y * g_width * 4
	if (y > 0) then	' above line exist
		n = y * (g_width * 4 + 1) + 1
		urn = (y - 1) * g_width * 4 
		lrn = y * g_width * 4
		for x = 0 to g_width - 1
			if (x) then	' left pixel exist. reference is (above d_pixel + left d_pixel) \ 2
				image_rgba[an] = cubyte((image_rgba[urn] + image_rgba[lrn])\2 + cbyte(image[n])): an+=1: urn+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte((image_rgba[urn] + image_rgba[lrn])\2 + cbyte(image[n])): an+=1: urn+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte((image_rgba[urn] + image_rgba[lrn])\2 + cbyte(image[n])): an+=1: urn+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte((image_rgba[urn] + image_rgba[lrn])\2 + cbyte(image[n])): an+=1: urn+=1: lrn+=1: n+=1
			else	' left pixel not exist. reference is above d_pixel \ 2
				image_rgba[an] = cubyte(image_rgba[urn]\2 + cbyte(image[n])): an+=1: urn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn]\2 + cbyte(image[n])): an+=1: urn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn]\2 + cbyte(image[n])): an+=1: urn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn]\2 + cbyte(image[n])): an+=1: urn+=1: n+=1
			endif
		next
	else	' above line not exist
		n = y * (g_width * 4 + 1) + 1
		lrn = y * g_width * 4
		for x = 0 to g_width - 1
			if (x) then	' left pixel exist. reference is left d-pixel \ 2
				image_rgba[an] = cubyte(image_rgba[lrn]\2 + cbyte(image[n])): an+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[lrn]\2 + cbyte(image[n])): an+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[lrn]\2 + cbyte(image[n])): an+=1: lrn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[lrn]\2 + cbyte(image[n])): an+=1: lrn+=1: n+=1
			else	' left pixel not exist.
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
			endif
		next
	end if	
end sub

' for paeth filter
function paethpredictor(a as ubyte, b as ubyte, c as ubyte) as ubyte
	dim as sint32 px, pa, pb, pc
	px = a + b - c
	pa = px - a: if (pa < 0) then invert_sign(pa)
	pb = px - b: if (pb < 0) then invert_sign(pb)
	pc = px - c: if (pc < 0) then invert_sign(pc)
	if ((pa <= pb) and (pa <= pc)) then
		return(a)
	else
		if (pb <= pc) then
			return(b)
		else
			return(c)
		end if
	end if
end function

sub decode_filter_paeth(image_rgba as ubyte ptr, image as ubyte ptr, y as sint32)
	dim as sint32 x, an, n, urn, rn
	dim as rgba_data c, c_ul, c_up, c_lf, paeth 

	an = y * g_width * 4
	if (y) then 	' above line exist
		urn = (y - 1) * g_width * 4
		rn = y * g_width * 4
		n = y * (g_width * 4 + 1) + 1
		for x = 0 to g_width - 1
			if (x) then	' left and above pixel exist (x>0, y>0). reference is calculated by paethpredictor
				c_ul.r = image_rgba[urn]: urn+=1
				c_ul.g = image_rgba[urn]: urn+=1
				c_ul.b = image_rgba[urn]: urn+=1
				c_ul.a = image_rgba[urn]: urn+=1
				c_up.r = image_rgba[urn] 
				c_up.g = image_rgba[urn + 1] 
				c_up.b = image_rgba[urn + 2] 
				c_up.a = image_rgba[urn + 3] 
				c_lf.r = image_rgba[rn]: rn+=1
				c_lf.g = image_rgba[rn]: rn+=1
				c_lf.b = image_rgba[rn]: rn+=1
				c_lf.a = image_rgba[rn]: rn+=1
				' get reference rgba value
				paeth.r = paethpredictor(c_lf.r, c_up.r, c_ul.r)
				paeth.g = paethpredictor(c_lf.g, c_up.g, c_ul.g)
				paeth.b = paethpredictor(c_lf.b, c_up.b, c_ul.b)
				paeth.a = paethpredictor(c_lf.a, c_up.a, c_ul.a)
				image_rgba[an] = cubyte(paeth.r + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(paeth.g + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(paeth.b + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(paeth.a + cbyte(image[n])): an+=1: n+=1
			else	' left pixel not exist so reference is above
				image_rgba[an] = cubyte(image_rgba[urn]     + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn + 1] + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn + 2] + cbyte(image[n])): an+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[urn + 3] + cbyte(image[n])): an+=1: n+=1
			end if
		next x
	else	'upper line not exist (x=*, y=0)
		rn = y * g_width * 4
		n = y * (g_width * 4 + 1) + 1
		for x = 0 to g_width - 1
			if (x) then	' left pixel is exist so reference pixel is left one.
				image_rgba[an] = cubyte(image_rgba[rn] + cbyte(image[n])): an+=1: rn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[rn] + cbyte(image[n])): an+=1: rn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[rn] + cbyte(image[n])): an+=1: rn+=1: n+=1
				image_rgba[an] = cubyte(image_rgba[rn] + cbyte(image[n])): an+=1: rn+=1: n+=1
			else	'left pixel is not exist (x=0, y=0)
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
				image_rgba[an] = image[n]: an+=1: n+=1
			end if
		next x
	end if
end sub

end namespace
