'-------------------------------------------------
'		sr2ctsmapper Saints Row 2 mapping Tool
'		Copyright (c) 2019 Masamichi Kirikae
'-------------------------------------------------
#include "sr2ctsmapper.bi"
	
sub zvalue_to_color(z as double, byref r as sint32, byref g as sint32, byref b as sint32)
	if z > 0 then
		r = z * 7
		if r > 255 then r = 255
		g = 255
		b = 0
	else
		r = 255
		g = 0
		b = -32 * z
		if b > 255 then b = 255
	endif
end sub

' convert rgba color to rgb color with white background. 
function rgba_to_rgb(c as rgba_data) as rgba_data
	dim t as double, tc as rgba_data
	t = c.r * (c.a / 255.0d) + 255.0d * (1.0d - c.a / 255.0d):	if t > 255.0d then t = 255.0d
	tc.r = cubyte(t)
	t = c.g * (c.a / 255.0d) + 255.0d * (1.0d - c.a / 255.0d):	if t > 255.0d then t = 255.0d
	tc.g = cubyte(t)
	t = c.b * (c.a / 255.0d) + 255.0d * (1.0d - c.a / 255.0d):	if t > 255.0d then t = 255.0d
	tc.b = cubyte(t)
	return tc	
end function

sub set_rgba(c as rgba_data, r as sint32, g as sint32, b as sint32)
	c.r = cubyte(r)
	c.g = cubyte(g)
	c.b = cubyte(b)
	c.a = 255
end sub

function readnavpoint(strline as string) as sint32
	dim as sint32 i, lg, stp, edp, n, flag
	dim as string ms, ds, cs, ms2, xs, ys, zs
	lg = len (strline)
	flag = FALSE
	for i = 1 to lg
		ms = mid(strline, i, 1)
		if ms = ":" then
			flag = TRUE
			exit for 
		endif
	next
	if (flag = FALSE) then 
		return(CTS_IGNORE)
	endif
	
	ms = mid(strline, 1, i)
	ds = mid(strline, i+1, lg - i)
	ds = ltrim(ds, " ")
	ds = ltrim(ds, chr(9))
	
	select case ms
		case "$Navpoint:" 
			navpts(navnum).navpoint = ds
			return(CTS_NAVPOINT)
		case "$Type:"
			navpts(navnum).pos_type = ds
			return(CTS_TYPE)
		case "$Pos:"
				lg = len(ds)
				cs = right(ds, lg - 1)
				n = 1
				xs = ""
				do
					ms2 = mid(cs, n, 1)
					n+=1
					if ms2 <> " " then
						xs = xs + ms2
					else
						exit do
					endif
				loop
				zs = ""
				do
					ms2 = mid(cs, n, 1)
					n+=1
					if ms2 <> " " then
						zs = zs + ms2
					else
						exit do
					endif
				loop
				ys = ""
				do
					ms2 = mid(cs, n, 1)
					n+=1
					if ms2 <> ">" then
						ys = ys + ms2
					else
						exit do
					endif
				loop
				navpts(navnum).pos_x = val(xs)
				navpts(navnum).pos_y = val(ys)
				navpts(navnum).pos_z = val(zs)
				return(CTS_POS)
		case "$Orient:"
				navpts(navnum).orient = ds
				return(CTS_ORIENT)
		case else
			return CTS_IGNORE
	end select
end function

sub graphic_line(image_rgba() as ubyte, xs as sint32, ys as sint32, xe as sint32, ye as sint32, c as rgba_data)
	dim as sint32 n, x, y
	if xs = xe then
		for y = ys to ye
			n = y * graphic_width * 4 + xs * 4
			image_rgba(n) = c.r: n+=1
			image_rgba(n) = c.g: n+=1
			image_rgba(n) = c.b: n+=1
			image_rgba(n) = c.a
		next
	else
		if ys = ye then
			for x = xs to xe
				n = ys * graphic_width * 4 + x * 4
				image_rgba(n) = c.r: n+=1
				image_rgba(n) = c.g: n+=1
				image_rgba(n) = c.b: n+=1
				image_rgba(n) = c.a				
			next
		endif
	endif
end sub

sub graph_box(image_rgba() as ubyte, xs as sint32, ys as sint32, xe as sint32, ye as sint32, c as rgba_data)
	dim as sint32 x, y
	for y = ys to ye
		graphic_line(image_rgba(), xs, y, xe, y, c)
	next
end sub

sub graphic_draw(image_rgba() as ubyte, xs as sint32, ys as sint32, ch as string)
	dim as sint32 x, y, gxs, gys, gxe, gye
	dim as rgba_data crgba
	dim as ubyte ptr bptr
	
	select case ch
		case "+"
			bptr = @ch_p(0, 0)
		case "-"
			bptr = @ch_m(0, 0)
		case "0"
			bptr = @ch_z(0, 0)
		case else
			bptr = @ch_p(0, 0)
	end select
	
	set_rgba(crgba, &HFF, &HFF, &HFF)	' White
	for y = 0 to 6
		for x = 0 to 6
			if *bptr <> 0 then
				gxs = xs + x * Char_Size
				gys = ys + y * Char_Size
				gxe = gxs + Char_Size - 1
				gye = gys + Char_Size - 1
				graph_box(image_rgba(), gxs, gys, gxe, gye, crgba)
			endif
			bptr+=1
		next
	next
end sub

sub drawcross(imgage_rgba() as ubyte, gx as sint32, gy as sint32, c as rgba_data)
	dim as sint32 gxs, gxe, gys, gye
	gxs = gx - Cross_Size
	gxe = gx + Cross_Size
	gys = gy - Cross_Size
	gye = gy + Cross_Size
	graphic_line(imgage_rgba(), gxs, gy, gxe, gy, c)
	graphic_line(imgage_rgba(), gx, gys, gx, gye, c)
end sub

sub drawmap_to_image(byref image as any ptr, image_rgba() as ubyte, zoomfactor as sint32)
	dim as sint32 x, y, n, r, g, b, a

	imagedestroy(image)
	image = imagecreate( 2048*zoomfactor, 2048*zoomfactor, RGBA(0, 0, 0, 255) )
	
	n = 0
	for y = 0 to 2048 * zoomfactor - 1 step zoomfactor
		for x = 0 to 2048 * zoomfactor - 1 step zoomfactor
			r = cast(sint32, image_rgba(n)):n+=1
			g = cast(sint32, image_rgba(n)):n+=1
			b = cast(sint32, image_rgba(n)):n+=1
			a = cast(sint32, image_rgba(n)):n+=1
			select case zoomfactor
				case 1
					pset image,(x, y), RGBA(r,g,b,a)
				case 2
					line image,(x, y)-(x + 1, y + 1), RGBA(r,g,b,a), BF
				case 4
					line image,(x, y)-(x + 3, y + 3), RGBA(r,g,b,a), BF					
				case else
			end select
		next
	next
end sub

function getmousestat(mousestat as mouse_data) as sint32
	dim as sint32 x, y, wheel, buttons, clip, retval
	static laststat as mouse_data
	dim as sint32 lb, rb, mb

	retval = GetMouse(x, y , wheel, buttons, clip)

	if retval <> 0 then return(1)
	if clip = 1 then return(1)
			
	lb = buttons and 1
	rb = (buttons shr 1) and 1
	mb = (buttons shr 2) and 1

	mousestat.x = x
	mousestat.y = y
	mousestat.wheel = wheel
	mousestat.lbutton = lb
	mousestat.rbutton = rb
	mousestat.mbutton = mb
	mousestat.clip = clip

	laststat = mousestat
	return(clip)	' return 0 when clipped
end function

'---------------------------------
'		main
'---------------------------------

function main() as sint32
	dim srcfilename as string, dstfilename as string, mapfile as string
	dim as sint32 imgw, imgh, bytepp
	dim as sint32 n, fi, gx, gy, r, g, b, a
	dim as double x, y, z
	dim stroneline as string
	dim as rgba_data crgba
	redim image_rgba(any) as ubyte 
	
	'	get source filename through comannd line parameter
	srcfilename = command(1)

	using FB
	screencontrol SET_DRIVER_NAME, "GDI"
	screenres WindowWidth, WindowHeight, 32,, GFX_HIGH_PRIORITY
	WindowTitle "SR2 ctsmapper"
	Width WindowWidth/8, WindowHeight/16
	
	if dir(srcfilename) = "" then
		print "Usage: sr2ctsmapper.exe <ctsfilename> or,"
		print "Drag *.cts file and drop it on sr2ctsmapper.exe"
		print ""
		print "press any key or close window."
		sleep
		return(0)
	end if
	
	dim extname as string
	extname = libfio.get_filename_ext(srcfilename)
	if extname <> "cts" then
		print "Error:Can not open this file."
		print ""
		print "press any key or close window."
		sleep
		return(0)
	endif
	
	mapfile = exepath() + uimapfilename
	dstfilename = libfio.change_filename_ext(srcfilename, ".png")

	print "Loading map...";	
	n = libpng.load_png(mapfile, image_rgba(), imgw, imgh, bytepp)

	print ""
	print "Processing...";

	graphic_width = imgw
	graphic_height = imgh
	
	' draw height color graph
	for gy = 0 to 511
		if gy < 256 then
			r = 255 - gy  
			g = 255
			b = 0
		else
			r = 255
			g = 0
			b = gy - 255
		endif
		set_rgba(crgba, r, g, b)
		graphic_line(image_rgba(), 0, gy, 32, gy, crgba)
	next
	graphic_draw(image_rgba(), 36,   8, "+")
	graphic_draw(image_rgba(), 36, 241, "0")
	graphic_draw(image_rgba(), 36, 472, "-")
	
	'	open cts file	
	fi = freefile
	open srcfilename for input as #fi
	
	navnum = 0
	do while(not(eof(fi)))
		line input #fi, stroneline
		n = readnavpoint(stroneline)
		select case n
			case CTS_POS
				x = navpts(navnum).pos_x
				y = navpts(navnum).pos_y
				z = navpts(navnum).pos_z

				'	convert to graphics coodinate
				gx = cint( ((-1.0 * x / 5808.0) * 2048.0) + 1256.0 )
				gy = cint( ((y / 5640.0) * 2048.0) + 1130.0 )	'1132.0

				navpts(navnum).pos_gx = gx
				navpts(navnum).pos_gy = gy
	
				zvalue_to_color(z, r, g, b)
				set_rgba(navpts(navnum).colour, r, g, b)
				drawcross(image_rgba(), gx, gy, navpts(navnum).colour)
			case CTS_ORIENT
				navnum+=1
			case else
		end select
	loop
	
	close fi
	
	libpng.save_png(dstfilename, image_rgba(), imgw, imgh, 32)
	
	cls
	
	dim as mouse_data mousestat, mousestatlast
	dim as sint32 zoomfactor = 1, MapSize = 2048
	dim image as any ptr = imagecreate( MapSize, MapSize, RGBA(0, 0, 0, 255) )

	drawmap_to_image(image, image_rgba(), zoomfactor)
	Window screen (0,  0) - (WindowWidth - 1, WindowHeight - 1)
	put(0, 0), image, pset

	dim as string kstr
	dim as sint32 retval, lastwheel, foundflag, printflag
	dim as sint32 ofsx, ofsy, ofsxlast, ofsylast
	dim as double gdx, gdy, distance
	
	ofsx = 0
	ofsy = 0
	ofsxlast = 0
	ofsylast = 0
	foundflag = FALSE
	printflag = FALSE
	do
		retval = getmousestat(mousestat)
		if (retval = 0) then
			if (mousestat.wheel - mousestatlast.wheel > 0) then			'	zoom in
				mousestatlast = mousestat
				if (zoomfactor < 4) then
					zoomfactor*=2
					MapSize*=2
					Window screen
					drawmap_to_image(image, image_rgba(), zoomfactor)
					ofsx = ofsx * 2 + mousestat.x
					ofsy = ofsy * 2 + mousestat.y
					if (ofsx + WindowWidth > MapSize - 1) then ofsx = MapSize - WindowWidth
					if (ofsy + WindowHeight > MapSize - 1) then ofsy = MapSize - WindowHeight
					Window screen (ofsx,  ofsy) - (WindowWidth + ofsx - 1, WindowHeight + ofsy - 1)
					put(0, 0), image, pset
				endif
			elseif (mousestat.wheel - mousestatlast.wheel < 0) then		'	zoom out
				mousestatlast = mousestat
				if (zoomfactor <= 4 and zoomfactor > 1) then
					zoomfactor/=2
					mapsize/=2
					Window screen
					drawmap_to_image(image, image_rgba(), zoomfactor)
					ofsx = ofsx / 2 - mousestat.x/2
					ofsy = ofsy / 2 - mousestat.y/2
					if (ofsx + WindowWidth > MapSize - 1) then ofsx = MapSize - WindowWidth
					if (ofsx < 0) then ofsx = 0
					if (ofsy + WindowHeight > MapSize - 1) then ofsy = MapSize - WindowHeight
					if (ofsy < 0) then ofsy = 0
					Window screen (ofsx,  ofsy) - (WindowWidth + ofsx - 1, WindowHeight + ofsy - 1)
					put(0, 0), image, pset
				endif				
			endif
			
			if (mousestat.mbutton = 1 or mousestat.rbutton = 1) then
				
				'	move viewport while mouse middle button pressing
			
				ofsx -= mousestatlast.x - mousestat.x
				if (ofsx + WindowWidth > MapSize) then
					ofsx = MapSize - WindowWidth
				elseif (ofsx < 0) then
					ofsx = 0
				endif 

				ofsy -= mousestatlast.y - mousestat.y
				if (ofsy + WindowHeight > MapSize) then
					ofsy = MapSize - WindowHeight
				elseif (ofsy < 0) then
					ofsy = 0
				endif
				
				if (ofsx <> ofsxlast or ofsy <> ofsylast) then
					Window screen (ofsx,  ofsy) - (WindowWidth + ofsx, WindowHeight + ofsy)
					screenlock
					put(0, 0), image, pset
					screenunlock
				endif				
				
			endif
			
			mousestatlast = mousestat

			gdx = (ofsx + mousestat.x)/zoomfactor
			gdy = (ofsy + mousestat.y)/zoomfactor

			foundflag = FALSE
			for n = 0 to navnum - 1
				distance = sqr( (gdx - navpts(n).pos_gx)^2.0 + (gdy - navpts(n).pos_gy)^2.0 )
				if (distance < 2.0) then
					foundflag = TRUE
					printflag = TRUE
					locate 45, 1
					'            0        1         2         3         4         5         6         7         8         9         10
					'				 12345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234
					print using "\                                      \ \       \ X:#####.###### Y:#####.###### Z:#####.###### Orient:\          \"; navpts(n).navpoint; navpts(n).pos_type; navpts(n).pos_x; navpts(n).pos_y; navpts(n).pos_z; navpts(n).orient;
					exit for
				endif
			next
			
			if (printflag = TRUE and foundflag = FALSE) then
				put(0, 0), image, pset
				printflag = FALSE
			endif
		endif
		kstr = inkey()
		select case kstr
			case chr(255) & "k"
				exit do
			case else
		end select
	loop

	return(0)
end function

end( main() )